import builtins
from datetime import datetime
import socket
import time
import threading

from bs4 import BeautifulSoup
from requests import Session

from dokuWikiDumper.exceptions import ActionEditDisabled, ActionEditTextareaNotFound, DispositionHeaderMissingError


from .revisions import getRevisions, getSourceEdit, getSourceExport
from .titles import getTitles
from dokuWikiDumper.utils.util import loadTitles, smkdirs, uopen
from dokuWikiDumper.utils.util import print_with_lock as print


sub_thread_error = None


def dumpContent(doku_url: str = '', dumpDir: str = '', session: Session = None, skipTo: int = 0,
                threads: int = 1, ignore_errors: bool = False, ignore_action_disabled_edit: bool = False,
                current_only: bool = False):
    if not dumpDir:
        raise ValueError('dumpDir must be set')

    titles = loadTitles(titlesFilePath=dumpDir + '/dumpMeta/titles.txt')
    if titles is None:
        titles = getTitles(url=doku_url, session=session)
        with uopen(dumpDir + '/dumpMeta/titles.txt', 'w') as f:
            f.write('\n'.join(titles))
            f.write('\n--END--\n')

    if not len(titles):
        print('Empty wiki')
        return False

    r1 = session.get(doku_url, params={'id': titles[0], 'do': 'export_raw'})
    r2 = session.get(doku_url, params={'id': titles[0]})
    r3 = session.get(doku_url, params={'id': titles[0], 'do': 'diff'})

    getSource = getSourceExport
    if 'html' in r1.headers['content-type']:
        print('\nWarning: export_raw action not available, using edit action\n')
        time.sleep(3)
        getSource = getSourceEdit

    soup = BeautifulSoup(r2.text, 'lxml')
    hidden_rev = soup.findAll(
        'input', {
            'type': 'hidden', 'name': 'rev', 'value': True})
    use_hidden_rev = hidden_rev and hidden_rev[0]['value']
    # TODO: what the `use_hidden_rev` is for?

    soup = BeautifulSoup(r3.text, 'lxml')
    select_revs = soup.findAll(
        'select', {
            'class': 'quickselect', 'name': 'rev2[0]'})
    # TODO: what the `select_revs` is for?

    index_of_title = -1  # 0-based
    if skipTo > 0:
        index_of_title = skipTo - 2
        titles = titles[skipTo-1:]

    def try_dump_page(*args, **kwargs):
        try:
            dump_page(*args, **kwargs)
        except Exception as e:
            if isinstance(e, ActionEditDisabled) and ignore_action_disabled_edit:
                print('[',args[2]+1,'] action disabled: edit. ignored')
            elif isinstance(e, ActionEditTextareaNotFound) and ignore_action_disabled_edit:
                print('[',args[2]+1,'] action edit: textarea not found. ignored')
            elif not ignore_errors:
                global sub_thread_error
                sub_thread_error = e
                raise e
            else:
                print('[',args[2]+1,']Error in sub thread: (', e, ') ignored')
    for title in titles:
        while threading.active_count() > threads:
            time.sleep(0.1)
        if sub_thread_error:
            raise sub_thread_error

        index_of_title += 1
        t = threading.Thread(target=try_dump_page, args=(dumpDir,
                                                     getSource,
                                                     index_of_title,
                                                     title,
                                                     doku_url,
                                                     session,
                                                     use_hidden_rev,
                                                     select_revs,
                                                     current_only,
                                                     ))
        print('Content: (%d/%d): [[%s]] ...' % (index_of_title+1, len(titles), title))
        t.daemon = True
        t.start()

    while threading.active_count() > 1:
        time.sleep(2)
        print('Waiting for %d threads to finish' %
              (threading.active_count() - 1), end='\r')


def dump_page(dumpDir: str,
              getSource,
              index_of_title: int,
              title: str,
              doku_url: str,
              session: Session,
              use_hidden_rev,
              select_revs,
              current_only: bool):
    srouce = getSource(doku_url, title, session=session)
    msg_header = '['+str(index_of_title + 1)+']: '
    child_path = title.replace(':', '/')
    child_path = child_path.lstrip('/')
    child_path = '/'.join(child_path.split('/')[:-1])

    smkdirs(dumpDir, '/pages/' + child_path)
    with uopen(dumpDir + '/pages/' + title.replace(':', '/') + '.txt', 'w') as f:
        f.write(srouce)

    if current_only:
        print(msg_header, '    [[%s]] saved.' % (title))
        return

    revs = getRevisions(doku_url, title, use_hidden_rev,
                        select_revs, session=session, msg_header=msg_header)

    revidOfPage: set[str] = set()
    for rev in revs[1:]:
        if 'id' in rev and rev['id']:
            try:
                txt = getSource(doku_url, title, rev['id'], session=session)
                smkdirs(dumpDir, '/attic/' + child_path)
                with uopen(dumpDir + '/attic/' + title.replace(':', '/') + '.' + rev['id'] + '.txt', 'w') as f:
                    f.write(txt)
                print(msg_header, '    Revision %s of [[%s]] saved.' % (
                    rev['id'], title))
            except DispositionHeaderMissingError:
                print(msg_header, '    Revision %s of [[%s]] is empty. (probably deleted)' % (
                    rev['id'], title))

            # time.sleep(1.5)

    smkdirs(dumpDir, '/meta/' + child_path)
    with uopen(dumpDir + '/meta/' + title.replace(':', '/') + '.changes', 'w') as f:
        # Loop through revisions in reverse.
        for rev in revs[::-1]:
            print(msg_header, '    meta change saving:', rev)
            sum = 'sum' in rev and rev['sum'].strip() or ''
            id = str(0)

            ip = '127.0.0.1'
            user = ''
            minor = 'minor' in rev and rev['minor']

            if 'id' in rev and rev['id']:
                id = rev['id']
            else:
                # Different date formats in different versions of DokuWiki.
                # If no ID was found, make one up based on the date (since rev IDs are Unix times)
                # Maybe this is evil. Not sure.

                print(
                    msg_header, '    One revision of [[%s]] missing rev_id. Using date to rebuild...' % title, end=' ')
                date_formats = ["%Y-%m-%d %H:%M", "%Y/%m/%d %H:%M", "%d.%m.%Y %H:%M", "%d/%m/%Y %H:%M"]
                # Try each date format until one works.
                # Example below:
                # %Y-%m-%d %H:%M | <https://www.dokuwiki.org/dokuwiki?do=revisions> # 2019/01/01 00:00
                # TODO: %Y-%m-%d %H:%M
                # TODO: %Y/%m/%d %H:%M
                # %d/%m/%Y %H:%M | <https://eolienne.f4jr.org/?do=revisions> #  28/02/2013 12:12
                for date_format in date_formats:
                    try:
                        date = datetime.strptime(rev['date'], date_format)
                        id = str(int(time.mktime(date.utctimetuple())))
                        break
                    except:
                        id = None
                       

                # if rev_id is not unique, plus 1 to it until it is.
                while id in revidOfPage:
                    id = str(int(id) + 1)
                print(msg_header, 'rev_id is now %s' % id)

            revidOfPage.add(id)

            rev['user'] = rev['user'] if 'user' in rev else 'unknown'
            try:
                # inet_aton throws an exception if its argument is not an IPv4 address
                socket.inet_aton(rev['user'])
                ip = rev['user']
            except socket.error:
                user = rev['user']

            sizechange = rev['sizechange'] if 'sizechange' in rev else ''

            extra = ''  # TODO: use this
            # max 255 chars(utf-8) for summary. (dokuwiki limitation)
            sum = sum[:255]
            row = '\t'.join([id, ip, 'e' if minor else 'E',
                            title, user, sum, extra, str(sizechange)])
            row = row.replace('\n', ' ')
            row = row.replace('\r', ' ')

            f.write((row + '\n'))





import urllib.parse as urlparse
from bs4 import BeautifulSoup
from dokuWikiDumper.utils.util import print_with_lock as print


def getTitles(url, ns=None, session=None):
    """Get titles given a doku.php URL and an (optional) namespace"""

    # # force use of old method for now
    # return getTitlesOld(url, ns=None, session=session)

    titles = []
    ajax = urlparse.urljoin(url, 'lib/exe/ajax.php')
    params = {'call': 'index'}
    if ns:
        params['idx'] = ns
    else:
        print('Finding titles')
    ns = ns or ''
    depth = len(ns.split(':'))
    if ns:
        print('%sLooking in namespace %s' % (' ' * depth, ns))
    r = session.post(ajax, params)
    if r.status_code != 200 or "AJAX call 'index' unknown!" in r.text:
        return getTitlesOld(url, ns=None, session=session)
    soup = BeautifulSoup(r.text, 'lxml')
    for a in soup.findAll('a', href=True):
        if a.has_attr('title'):
            title = a['title']
        else:
            query = urlparse.parse_qs(urlparse.urlparse(a['href']).query)
            title = (query['idx' if 'idx' in query else 'id'])[0]
        if 'idx_dir' in a['class']:
            titles += getTitles(url=url, ns=title, session=session)
        else:
            titles.append(title)
    # time.sleep(1.5)
    print('%sFound %d title(s) in namespace %s' %
          (' ' * depth, len(titles), ns or '(all)'))
    return titles


def getTitlesOld(url, ns=None, ancient=False, session=None):
    """Get titles using the doku.php?do=index"""

    titles = []
    params = {'do': 'index'}

    if ns:
        params['idx'] = ns
    ns = ns or ''
    depth = len(ns.split(':'))

    r = session.get(url, params=params)
    soup = BeautifulSoup(r.text, 'lxml').findAll('ul', {'class': 'idx'})[0]
    attr = 'text' if ancient else 'title'

    if ns:
        print('%sSearching in namespace %s' % (' ' * depth, ns))

        def match(href):
            if not href:
                return False
            qs = urlparse.urlparse(href).query
            qs = urlparse.parse_qs(qs)
            return 'idx' in qs and qs['idx'][0] in (ns, ':' + ns)
        result = soup.findAll(
            'a', {
                'class': 'idx_dir', 'href': match})[0].findAllPrevious('li')[0].findAll(
            'a', {
                'href': lambda x: x and not match(x)})
    else:
        print('Finding titles (?do=index)')
        result = soup.findAll('a')

    for a in result:
        query = urlparse.parse_qs(urlparse.urlparse(a['href']).query)
        if (
            a.has_attr('class')
            and ('idx_dir' in a['class'])
        ):
            titles += getTitlesOld(url, query['idx'][0], session=session)
        else:
            titles.append(query['id'][0])

    print('%sFound %d title(s) in namespace %s' %
          (' ' * depth, len(titles), ns or '(all)'))

    return titles



import argparse
import hashlib
import time
import os
import subprocess
import json

from internetarchive import get_item

from dokuWikiDumper.utils.util import url2prefix
from dokuWikiDumper.dump.info import get_info
from dokuWikiDumper.dump.info import INFO_WIKI_NAME, INFO_RAW_TITLE, INFO_DOKU_URL, INFO_LANG, INFO_ICON_URL
from dokuWikiDumper.utils.config import get_config

from .__version__ import UPLOADER_VERSION

DEFAULT_COLLECTION = 'opensource'
TEST_COLLECTION = 'test_collection' # items here are expected to be automatically removed after 30 days. 
# (see <https://archive.org/details/test_collection?tab=about>)

USER_AGENT = 'dokuWikiUploader/' + UPLOADER_VERSION

UPLOADED_MARK = 'uploaded_to_IA.mark'


def file_sha1(path):
    buffer = bytearray(65536)
    view = memoryview(buffer)
    digest = hashlib.sha1()

    with open(path, mode="rb") as f:
        while True:
            n = f.readinto(buffer)

            if not n:
                break

            digest.update(view[:n])

    return digest.hexdigest()


def read_ia_keys(keysfile):
    ''' Return: tuple(`access_key`, `secret_key`) '''
    with open(keysfile, 'r', encoding='utf-8') as f:
        key_lines = f.readlines()

    access_key = key_lines[0].strip()
    secret_key = key_lines[1].strip()

    return access_key, secret_key


def upload(args={}):
    dump_dir = args.dump_dir
    path7z = args.path7z # '/usr/bin/7z'
    access_key, secret_key = read_ia_keys(os.path.expanduser(args.keysfile))
    collection = args.collection
    pack_dumpMeta_dir = args.pack_dumpMeta
    info = get_info(dump_dir)
    config = get_config(dump_dir)
    headers = {"User-Agent": USER_AGENT}


    title = (info.get(INFO_WIKI_NAME) if info.get(INFO_WIKI_NAME) else info.get(INFO_RAW_TITLE))
    title = title if info.get(INFO_RAW_TITLE) else url2prefix(info.get(INFO_DOKU_URL))
    title = title if title else 'Unknown'
    try:
        _dump_dir_ab = os.path.abspath(dump_dir)
        _dump_dir_basename = os.path.basename(_dump_dir_ab)
        __dump_dir_basename = _dump_dir_basename.split('-')
        if len(__dump_dir_basename) >= 2: # Punycoded domain may contain multiple `-`
            if int(__dump_dir_basename[-1]) < 20230200:
                raise ValueError('Invalid dump directory name: %s, OMG, I cannot believe it was created before dokuWikiDumper was born!? (0w0)' % _dump_dir_basename)
        else:
            raise ValueError('Invalid dump directory name: %s' % _dump_dir_basename)
        identifier_local = _dump_dir_basename
    except Exception as e:
        print(e)
        print("\nSomething went wrong during generating the identifier. \n")
        raise e
    
    identifier_remote = 'wiki-' + identifier_local # 'wiki-' is the prefix of all wikis in archive.org
    # with the prefix, wikidump can go into wikiteam collection.

    # identifier_remote = 'test-wiki-' + identifier_local # 

    description_with_URL = \
f'''DokuWiki: <a href="{info.get(INFO_DOKU_URL)}" rel="nofollow">{title}</a>
<br>
<br>
Dumped with <a href="https://github.com/saveweb/dokuwiki-dumper" rel="nofollow">DokuWiki-Dumper</a> v{config.get('dokuWikiDumper_version')}, and uploaded with dokuWikiUploader v{UPLOADER_VERSION}.'''
    description_without_URL = \
f'''DokuWiki: {title}
<br>
<br>
Dumped with DokuWiki-Dumper v{config.get('dokuWikiDumper_version')}, and uploaded with dokuWikiUploader v{UPLOADER_VERSION}.'''
    keywords_init = [
        "wiki",
        "wikiteam",
        "DokuWiki",
        "dokuWikiDumper",
        "wikidump",
    ]
    keywords_full = keywords_init.copy()
    if title and title not in keywords_full:
        keywords_full.append(title)
    if info.get(INFO_DOKU_URL):
        keywords_full.append(url2prefix(info.get(INFO_DOKU_URL)))


    # Item metadata
    md_init = {
        "mediatype": "web",
        "collection": collection,
        "title": "Wiki - " + title,
        "description": description_without_URL, # without URL, to bypass IA's anti-spam.
        "last-updated-date": time.strftime("%Y-%m-%d"),
        "subject": "; ".join(
            keywords_init
        ),  # Keywords should be separated by ; but it doesn't matter much; the alternative is to set one per field with subject[0], subject[1], ...
    }
    language = info.get(INFO_LANG) if info.get(INFO_LANG) else config.get('lang')
    if language:
        md_init.update({
            "language": language,
        })


    dirs_to_7z = ["attic","html","media","pages", "pdf"]
    mark_files = {"attic":  "content_dumped.mark", 
                "pages":    "content_dumped.mark",
                "html":     "html_dumped.mark",
                "media":    "media_dumped.mark",
                "pdf":      "pdf_dumped.mark",
                "dumpMeta": "dumpMeta/" # no .mark file for dumpMeta, check itself instead.
                }
    filedict = {} # "remote filename": "local filename"

    if pack_dumpMeta_dir:
        dirs_to_7z.append("dumpMeta")
    else:
        # list all files in dump_dir/dumpMeta
        for dir in os.listdir(os.path.join(dump_dir, "dumpMeta/")):
            filedict.update({identifier_local+"-"+"dumpMeta/"+os.path.basename(dir): os.path.join(dump_dir, "dumpMeta/", dir)})

    for dir in dirs_to_7z:
        _dir = os.path.join(dump_dir, dir)
        if os.path.isdir(_dir):

            if not os.path.exists(os.path.join(dump_dir, mark_files[dir])):
                raise Exception(f"Directory {dir} is not finished. Please run dokuWikiDumper again. ({mark_files[dir]} not found)")

            print(f"Compressing {_dir}...")
            level = 1 if (dir == "media" or dir == "pdf") else 5
            filedict.update({identifier_local+"-"+dir+".7z": compress(_dir, path7z, level=level)})

    # Upload files and update metadata
    print("Preparing to upload...")
    try:
        print(f"Identifier (Local): {identifier_local}")
        print(f"Identifier (Remote): {identifier_remote}")
        item = get_item(identifier_remote)
        for file_in_item in item.files:
            if file_in_item["name"] in filedict:
                filedict.pop(file_in_item["name"])
                print(f"File {file_in_item['name']} already exists in {identifier_remote}.")
        print(f"Uploading {len(filedict)} files...")
        print(md_init)
        r = item.upload(
            files=filedict,
            metadata=md_init,
            access_key=access_key,
            secret_key=secret_key,
            verbose=True,
            queue_derive=False,
        )
        print(f"Uploading {len(filedict)} files: Done.\n")


        tries = 20
        while not item.exists and tries > 0:
            print("Waiting for item to be created...", tries)
            tries -= 1
            time.sleep(10)
            item = get_item(identifier_remote)



        print("Updating description...")
        new_md  = {}
        if (info.get(INFO_DOKU_URL) not in item.metadata.get("description") or
            'https://github.com/saveweb/dokuwiki-dumper' not in item.metadata.get("description")):
            # IA will format the description's HTML, so we can't just use `==` to compare.
            print("    (add URL back)...")
            new_md.update({"description": description_with_URL})
        if item.metadata.get("last-updated-date") != time.strftime("%Y-%m-%d"):
            print("    (update last-updated-date)...")
            new_md.update({"last-updated-date": time.strftime("%Y-%m-%d")})
        if item.metadata.get("subject") != "; ".join(keywords_full):
            print("    (update subject)...")
            new_md.update({"subject": "; ".join(keywords_full)})
        if item.metadata.get("originalurl") != info.get(INFO_DOKU_URL):
            print("    (update originalurl)...")
            new_md.update({"originalurl": info.get(INFO_DOKU_URL)})

        if new_md:
            r = item.modify_metadata(metadata=new_md,  # update
                                    access_key=access_key, secret_key=secret_key)
            print(r.text)
            r.raise_for_status()
            print("Updating description: Done.")
        else:
            print("Updating description: No need to update.")


        print(
            "\n\n--Done--\nYou can find it in https://archive.org/details/%s"
            % (identifier_remote)
        )
    except Exception as e:
        raise e

def compress(dir, bin7z: str, level: int = 5):
    ''' Compress dir into dump_dir.7z and return the absolute path to the compressed file. '''
    dir = os.path.abspath(dir) # remove trailing slash
    dir_7z = dir + ".7z"
    if os.path.exists(dir_7z):
        print(f"File {dir_7z} already exists. Skip compressing.")
        return dir_7z
    
    subprocess.run(
        [bin7z, "a", "-t7z", "-m0=lzma2", f"-mx={level}", "-scsUTF-8",
            "-md=64m", "-ms=off", dir_7z+".tmp", dir],
        check=True,
    )
    # move tmp file to final file
    os.rename(dir_7z+".tmp", dir_7z)
    return dir_7z


def main(params=[]):

    parser = argparse.ArgumentParser(
        """dokuWikiUploader"""
    )
    parser.description = "Upload a DokuWiki dump to Internet Archive." + f" (Version: {UPLOADER_VERSION})."
    parser.add_argument("-kf", "--keysfile", default="~/.doku_uploader_ia_keys",
                        help="Path to the IA S3 keys file. (first line: access key, second line: secret key)"
                             " [default: ~/.doku_uploader_ia_keys]")
    parser.add_argument("-p7z", "--path7z", default="7z",
                        help="Path to 7z binary. [default: 7z]")
    parser.add_argument("-c", "--collection", default=DEFAULT_COLLECTION, choices=[DEFAULT_COLLECTION, TEST_COLLECTION, "wikiteam"],
                        help="Collection to upload to. ('test_collection' for testing (auto-delete after 30 days) "
                             "[default: opensource]")
    parser.add_argument("-p", "--pack-dumpMeta", action="store_true",
                        help="Pack the dumpMeta/ directory into a 7z file, then upload it. "
                             "instead of uploading all files in dumpMeta/ directory individually. "
                             "[default: False]")
    parser.add_argument("dump_dir", help="Path to the wiki dump directory.")
    args = parser.parse_args()

    if os.path.exists(os.path.join(args.dump_dir, UPLOADED_MARK)):
        print("This dump has already been uploaded.")
        print("If you want to upload it again, please remove the file '%s'." % UPLOADED_MARK)
        return 0

    upload(args)

    with open(os.path.join(args.dump_dir, UPLOADED_MARK), "w", encoding='UTF-8') as f:
        f.write("Uploaded to Internet Archive with dokuWikiUploader v%s on %s" % (UPLOADER_VERSION, time.strftime("%Y-%m-%d %H:%M:%S")))

if __name__ == "__main__":
    main()
